__doc__ = """
For the terminal.
"""

from logging import INFO, basicConfig, getLogger
from sys import exit

from click import Tuple, command, option
from rich.console import Console
from rich.logging import RichHandler

from .logo import Logo, LogoParams

console = Console()
basicConfig(
    level=INFO,
    format="%(message)s",
    datefmt=None,
    handlers=[
        RichHandler(
            console=console,
            markup=True,
            rich_tracebacks=True,
            show_path=True,
            enable_link_path=False,
            show_time=False,
        )
    ],
)
logger = getLogger(__name__)

filename_default = "Matches the 'text' parameter by default"


@command()
@option(
    "--text",
    type=str,
    required=True,
    prompt="The text to add to your logo",
    help="The text to add to your logo",
)
@option(
    "--font",
    type=str,
    required=False,
    help="The ASCII font that you want to use in your logo",
    default=LogoParams.font,
)
@option(
    "--fontsize",
    type=int,
    required=False,
    help="The size of the characters that make the ANSII image into your logo",
    default=LogoParams.font_size,
)
@option(
    "--fontcolor",
    type=Tuple([int, int, int, int]),
    required=False,
    help="RGBA value for the font",
    default=LogoParams.font_color,
)
@option(
    "--backgroundcolor",
    type=Tuple([int, int, int, int]),
    required=False,
    help="RGBA for the backgroun",
    default=LogoParams.background_color,
)
@option(
    "--filename",
    type=str,
    required=False,
    help="The filename for the logo image",
    default=filename_default,
)
@option(
    "--fontpath",
    type=str,
    required=False,
    help="The path to the font for assembling your logo",
    default=LogoParams.font_path,
)
def create(
    text: str,
    font: str,
    fontsize: int,
    fontcolor: tuple,
    fontpath: str,
    backgroundcolor: tuple,
    filename: str,
):
    """Creates a logo using ASCII art."""

    if filename == filename_default:
        filename = LogoParams.filename
    try:
        Logo(
            text=text,
            font=font,
            font_size=int(fontsize),
            font_color=fontcolor,
            font_path=fontpath,
            background_color=backgroundcolor,
            filename=filename,
        ).create()
    except Exception as e:
        logger.error(f"[bold red]{type(e).__name__}[/bold red] : {e}")
        exit(1)



__all__ = ["InvalidRGBAError", "InvalidFontError"]

__doc__ = """
For when ya done goofed.
"""

from typing import Optional


class BaseError(Exception):
    """The mother of all exceptions."""

    def __init__(self, message: Optional[str] = None):
        self.message = "" if message is None else message
        super().__init__(message)

    def __str__(self) -> str:
        return self.message


class InvalidRGBAError(BaseError):
    """Raised when you provide an illegal RGBA value."""

    ...


class InvalidFontError(BaseError):
    """Raised when you provide an ASCII font that cannot be found in `art's font gallery <https://www.ascii-art.site/FontList.html>`_."""

    ...



__all__ = ["Logo"]

__doc__ = """
For creating a logo.
"""

from logging import getLogger
from pathlib import Path
from typing import Tuple

from art import text2art
from attrs import define, field
from attrs.validators import instance_of, optional
from PIL import Image, ImageDraw, ImageFont

from .validators import *

logger = getLogger(__name__)


class LogoParams:
    """Stores default values for the :class:`Logo` object.
    
    Attributes
    ----------
    font : str, optional
        The `ASCII font <https://www.ascii-art.site/FontList.html>`_ that you want to use in your logo. Default is 'alpha'.
    font_size : int, optional
        The size of the text characters that assemble the ASCII image for your logo. Default is 20.
    font_path : str, optional
        The path to the font file for the text characters that assemble the ASCII image for your logo.
    font_color : tuple, optional
        RGBA value for the text characters that assemble the ASCII image for your logo. Default is (255, 0, 0, 255)
    background_color : tuple, optional
        RGBA value for the background color of your logo. Default is (0, 0, 0, 0)
    filename : str, optional
        The location where you want your logo saved. Default matches the `text` parameter plus the ``.pn
    extension : str, optional
        The file extension. Default is '.png'.
    """

    font: str = "alpha"
    font_size: int = 20
    font_color: Tuple[int, int, int, int] = (255, 0, 0, 255)
    background_color: Tuple[int, int, int, int] = (0, 0, 0, 0)
    filename: str = ""
    extension: str = ".png"
    font_path: str = str(
        Path(__file__).parent / "fonts" / "./DejaVuSansMono.ttf"
    )


@define
class Logo:
    """Creates a logo using ASCII art.
    
    Attributes
    ----------
    text : str
        The text to add into your logo.
    font : str, optional
        The `ASCII font <https://www.ascii-art.site/FontList.html>`_ that you want to use in your logo. Default is 'alpha'.
    font_size : int, optional
        The size of the text characters that assemble the ASCII image for your logo. Default is 20.
    font_path : str, optional
        The path to the font file for the text characters that assemble the ASCII image for your logo.
    font_color : tuple, optional
        RGBA value for the text characters that assemble the ASCII image for your logo. Default is (255, 0, 0, 255)
    background_color : tuple, optional
        RGBA value for the background color of your logo. Default is (0, 0, 0, 0)
    filename : str, optional
        The location where you want your logo saved. Default matches the `text` parameter plus the ``.png`` extension.
    
    Methods
    -------
    create()
        Creates the logo.
    
    Notes
    -----
    Only PNG images are permitted in order to support the creation of images with transparent backgrounds via RGBA values.
    The size of images is relative to the size of the rendered ASCII art and cannot be controlled.
    Every font in the `art font gallery <https://www.ascii-art.site/FontList.html>`_ is available.
    The default character font file can be found in the fonts/ submodule of this package.
    
    See Also
    --------
    laudanum.exceptions.InvalidRGBAError
    laudanum.exceptions.InvalidFontError
    
    Examples
    --------
    >>> logo = Logo('howdy')
    >>> logo.create()
    """

    text: str = field(validator=[instance_of(str)])
    font: str = field(
        validator=optional([instance_of(str), validate_font]),
        default=LogoParams.font,
    )
    font_path: str = field(
        validator=optional([instance_of(str)]),
        default=LogoParams.font_path,
    )
    font_size: int = field(validator=optional([instance_of(int)]), default=20)
    font_color: Tuple[int, int, int, int] = field(
        validator=optional([instance_of(tuple), validate_rgba]),
        default=LogoParams.font_color,
    )
    background_color: Tuple[int, int, int, int] = field(
        validator=optional([instance_of(tuple), validate_rgba]),
        default=LogoParams.background_color,
    )
    filename: str = field(
        validator=optional([instance_of(str)]), default=LogoParams.filename
    )
    extension: str = field(init=False)
    extensions: dict = field(init=False)

    def __attrs_post_init__(self):

        self.extension = LogoParams.extension

        if self.filename:
            if "." in self.filename and not self.filename.endswith(
                self.extension
            ):
                logger.warning(f"Filenames must end in '{self.extension}'.")
            else:
                self.filename = self.filename.split(".")[0] + self.extension
        else:
            self.filename = f"{self.text}{self.extension}"

        self.filename = f"{self.text if not self.filename else self.filename.split('.')[0]}{self.extension}"
        self.extensions = Image.registered_extensions()

    def create(self):
        """Creates the ASCII logo.
        
        Raises
        ------
        InvalidRGBAError
            Raised when you provide an illegal RGBA value.
        InvalidFontError
            Raised when you provide a font that cannot be found in art's gallery.
        """

        ascii = text2art(self.text, font=self.font).splitlines()
        _font = ImageFont.truetype(self.font_path, self.font_size)
        bbox = _font.getbbox("A")
        img_width = (bbox[2] - bbox[0]) * max(len(line) for line in ascii)
        img_height = (char_height := bbox[3] - bbox[1]) * len(ascii)
        image = Image.new(
            "RGBA", (img_width, img_height), self.background_color
        )
        draw = ImageDraw.Draw(image)

        for i, line in enumerate(ascii):
            draw.text(
                (0, i * char_height), line, font=_font, fill=self.font_color
            )

        image.save(self.filename, self.extensions[self.extension])

        logger.info(f"Image saved at {self.filename}")



__all__ = ["validate_font", "validate_rgba"]

from typing import Tuple, Type

from art import FONT_NAMES
from attrs import Attribute

from .exceptions import *


def validate_font(
    instance: Type["Logo"], attribute: Type[Attribute], value: str
) -> None:
    if value not in FONT_NAMES:
        msg = f"Invalid font selected: '{value}'. Valid options can be found here: https://www.ascii-art.site/FontList.html"
        raise InvalidFontError(msg)


def validate_rgba(
    instance: Type["Logo"],
    attribute: Type[Attribute],
    value: Tuple[int, int, int, int],
):
    if min(value) < 0 or max(value) > 255 or len(value) != 4:
        msg = f"A valid RGBA value must contain three values not exceeding 255 or below 0. You provided '{value}'"
        raise InvalidRGBAError(msg) from None
